#!/usr/bin/perl -w
#
#  Calendar utility routines
#  Use this module to manipulate dates
#
#  Copyright (c) 2001  Willem A. Schreuder <willem@prinmath.com>
#  This file is part of the KUDU suite of software.
#
#  The KUDU programs are free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License as
#  published by the Free Software Foundation; either version 2 of the
#  License, or (at your option) any later version.
#
#  The KUDU programs are distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  General Public License for more details.
#
#  You should have received a copy of the GNU General Public License along 
#  with this program; if not, write to the Free Software Foundation, Inc.,
#  51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
#
#  See also www.fsf.org

use strict;
use POSIX;

#
#  Hash of month names
#
my %monthhash = (jan => 1 , feb =>  2 , mar =>  3 , apr =>  4,
                 may => 5 , jun =>  6 , jul =>  7 , aug =>  8,
                 sep => 9 , oct => 10 , nov => 11 , dec => 12);
my @monthname = (undef,'January','February','March','April','May','June','July','August','September','October','November','December');

#
#  Floor for integer division
#
sub Floor
{
  my ($x,$y) = @_;
  return POSIX::floor($x/$y);
}

#
#  Leap year
#
sub LeapYear
{
   my ($year,$cal) = @_;
   defined($cal) || ($cal = ($year>1582 ? 'G' : 'J'));
   return ($cal eq 'J') ? ($year%4==0) : ($year%4==0 && ($year%400==0 || $year%100!=0));
}

#
#  Real date
#
sub RealDate
{
   my ($year,$month,$day,$f,$cal) = @_;
   my @ndy = (undef,-1,30,58,89,119,150,180,211,242,272,303,333);
   #  If cal is undefined, dates from 10/15/1582 (JD 2299161) are assumed to be Gregorian Dates
   #  Earlier dates are assumed to be Julian Dates
   unless (defined $cal)
   {
      $cal = ($year>1582 || $year==1582 && $month>10 || $year==1582 && $month==10 && $day>=15) ? 'G' : 'J';
   }
   else
   {
      $cal eq 'J' || $cal eq 'G' || die "Calendar is not Julian or Gregorian $cal\n";
   }
   my $leap = LeapYear($year,$cal);
   my $ndy = $leap ? 366 : 365;
   my $doy = $ndy[$month] + $day + (($leap && $month>2) ? 1 : 0);
   defined($f) && ($doy += $f);
   return $year + $doy/$ndy;
}
 
#
#  Compute Julian day:  Number of days since 1-Jan-4713 BC (-4712-1-1)
#
sub JulianDay
{
   my ($year,$month,$day,$cal) = @_;
   my @MDO=(-1,306,337,0,31,61,92,122,153,184,214,245,275);
   #  If cal is undefined, dates from 10/15/1582 (JD 2299161) are assumed to be Gregorian Dates
   #  Earlier dates are assumed to be Julian Dates
   unless (defined $cal)
   {
      $cal = ($year>1582 || $year==1582 && $month>10 || $year==1582 && $month==10 && $day>=15) ? 'G' : 'J';
   }
   else
   {
      $cal eq 'J' || $cal eq 'G' || die "Calendar is not Julian or Gregorian $cal\n";
   }

   my $Y = ($month<3) ? $year-1 : $year;
   my $JD = $day+$MDO[$month]+365*$Y+Floor($Y,4)+1721117;
   if ($cal eq 'G')
   {
      $JD += 2-Floor($Y,100)+Floor($Y,400);
   }
   return $JD;
}
 
#
#  Name of Month
#
sub MonthName
{
   my ($mo,$long) = @_;
   return defined($monthname[$mo]) ? ($long?$monthname[$mo]:substr($monthname[$mo],0,3)) : '';
}
 
#
#  Compute weekday
#
sub WeekDay
{
   my @day = ('Mon','Tue','Wed','Thu','Fri','Sat','Sun');
   return $day[JulianDay(@_)%7];
}
 
#
#  Julian day from mm/dd/yyyy
#
sub JulianUS
{
   my ($date,$cal) = @_;
   ($date =~ /^(\d+)\/(\d+)\/(\d+)$/) || return;
   my ($mo,$dy,$yr) = ($1,$2,$3);
   return JulianDay($yr,$mo,$dy,$cal);
}
 
#
#  Julian day from YYYYMMDD
#
sub JulianYYYYMMDD
{
   my ($date,$cal) = @_;
   my ($y,$m,$d,$f) = YYYYMMDD($date);
   return JulianDay($y,$m,$d,$cal)+$f;
}
 
#
#  Compute Day of Year: 0-364 or 0-365 for leap years
#
sub DayOfYear
{
   my ($year,$month,$day,$cal) = @_;
   my $JD0 = JulianDay($year,1,1,$cal);
   my $JD1 = JulianDay($year,$month,$day,$cal);
   return $JD1-$JD0;
}

#
#  YYYYMMDDhhmmss
#
sub YYYYMMDD
{
   my ($date) = @_;
   my $yr = substr($date,0,4);
   my $mo = (length($date)>4)  ? substr($date,4,2)  : 0;
   my $dy = (length($date)>6)  ? substr($date,6,2)  : 0;
   my $hr = (length($date)>8)  ? substr($date,8,2)  : 0;
   my $mn = (length($date)>10) ? substr($date,10,2) : 0;
   my $sc = (length($date)>12) ? substr($date,12) : 0;
   ($dy ne '' && $dy>0) || ($dy = 15);
   ($mo>0) || (($mo,$dy) = (7,1));
   my $f = $hr/24+$mn/1440+$sc/86400;
   return ($yr,$mo,$dy,$f);
}

#
#  ddMMMyyyy
#
sub ddMMMyyyy
{
   my ($date) = @_;
   my ($dy,$mo,$yr) = split('-',$date,3);
   $mo = lc $mo;
   exists($monthhash{$mo}) || die "Invalid month $mo in $date\n";
   $mo = $monthhash{$mo};
   ($yr<35) && ($yr += 2000);
   ($yr<100) && ($yr += 1900);
   return ($yr,$mo,$dy);
}

#
#  Compute Calendar day
#
sub CalendarDay
{
   my ($JD,$cal) = @_;
   my @MDO=(-1,306,337,0,31,61,92,122,153,184,214,245,275);
   $JD = int($JD);
   #  If not defined cal, dates from 10/15/1582 (JD 2299161) are assumed to be Gregorian Dates
   #  Earlier dates are assumed to be Julian Dates
   unless (defined $cal)
   {
      $cal = ($JD>=2299161) ? "G" : "J";
   }
   elsif (uc $cal eq 'EXCEL')
   {
      $cal = 'G';
      $JD += 2415019;
   }
   else
   {
      $cal eq 'J' || $cal eq 'G' || die "Calendar is not Julian or Gregorian $cal";
   }

   my ($Z,$B,$H,$A);
   if ($cal eq 'G')
   {
      $Z = $JD-1721119;
      $H = 100*$Z-25;
      $A = Floor($H,3652425);
      $B = $A - Floor($A,4);
   }
   else
   {
      $Z = $JD-1721117;
      $H = 100*$Z-25;
      $B = 0;
   }
   my ($year,$month,$day);
   $year  = Floor(100*$B+$H,36525);
   $day   = $B+$Z-365*$year-Floor($year,4);
   $month = int((5*$day+456)/153);
   if ($month>12)
   {
      $month -= 12;
      $year++;
   }
   $day -= $MDO[$month];
   return ($year,$month,$day);
}

#
#  Number of days in a month
#
sub DaysInMonth
{
   my ($yr,$mo) = @_;
   exists($monthhash{lc $mo}) && ($mo = $monthhash{lc $mo});
   ($mo<1 || $mo>12) && die "Month out of range $mo\n";
   my @ndm = (undef,31,28,31,30,31,30,31,31,30,31,30,31);
   my $ndm;
   if (defined($yr))
   {
      $ndm = ($yr%4==0 && $mo==2) ? 29 : $ndm[$mo];
   }
   else
   {
      $ndm = ($mo==2) ? 28.25 : $ndm[$mo];
   }
   return $ndm;
}

#
#  Number of days in a year
#
sub DaysInYear
{
   my ($yr) = @_;
   return ($yr%4==0) ? 366 : 365;
}

#
#  UNIX Time (seconds since 1-Jan-1970)
#
sub unixtime
{
   my ($yr,$mo,$dy,$hr,$mn,$sc) = @_;
   return  86400*(JulianDay($yr,$mo,$dy)-2440588)+3600*$hr+60*$mn+$sc;
}

#
#  Local Standard Time
#
sub LST
{
   my $time = time();
   my ($sec,$min,$hour,$day,$mon,$year,$wday,$yday,$isdst) = localtime($time);
   ($isdst) && (($sec,$min,$hour,$day,$mon,$year,$wday,$yday,$isdst) = localtime($time-3600));
   return  ($year+1900,$mon+1,$day,$hour,$min,$sec);
}

#
#  Previous day
#
sub PreviousDay
{
   my ($yr,$mo,$dy) = @_;
   if ($dy>1)
   {
      $dy--;
   }
   else
   {
      if ($mo>1)
      {
         $mo--;
      }
      else
      {
         $mo = 12;
         $yr--;
      }
      $dy = DaysInMonth($yr,$mo);
   }
   return ($yr,$mo,$dy);
}

#
#  Next day
#
sub NextDay
{
   my ($yr,$mo,$dy) = @_;
   if ($dy<DaysInMonth($yr,$mo))
   {
      $dy++;
   }
   else
   {
      $dy = 1;
      if ($mo<12)
      {
         $mo++;
      }
      else
      {
         $mo = 1;
         $yr++;
      }
   }
   return ($yr,$mo,$dy);
}

#
#  Current time
#
sub now
{
   my ($sec,$min,$hour,$day,$mon,$year,$wday) = localtime();
   $year += 1900;
   $mon++;
   my $date = sprintf '%.4d-%.2d-%.2d' , $year , $mon , $day;
   my $time = sprintf '%.2d:%.2d:%.2d' , $hour , $min , $sec;
   return ($date,$time);
}

#  Sanity check
(JulianDay(CalendarDay(2440588)) == 2440588) || die "1-Jan-1970 failed Julian Day sanity check";
