#!/usr/bin/perl -w
#
#  FORTRAN Object
#  Use this module to read and write FORTRAN files
#
#  Copyright (c) 2001  Willem A. Schreuder <willem@prinmath.com>
#  This file is part of the KUDU suite of software.
#
#  The KUDU programs are free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License as
#  published by the Free Software Foundation; either version 2 of the
#  License, or (at your option) any later version.
#
#  The KUDU programs are distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  General Public License for more details.
#
#  You should have received a copy of the GNU General Public License along 
#  with this program; if not, write to the Free Software Foundation, Inc.,
#  51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
#
#  See also www.fsf.org

package FORTRAN;

use strict;
use FileHandle;
use IO::Seekable;

#
#  Open file
#
sub open
{
   my $class = shift;
   my $self = {};
   bless($self,$class);

   my ($file,$type) = @_;
   defined($type) || ($type = '4');
   ($type =~ /[048L]/) || die "Unknown open file type $type\n";

   my $fh = new FileHandle($file);
   defined($fh) || return;
   binmode($fh);
   $self->{WR} = ($file =~ /^ *>/ || $file =~ /^ *|/);
   $self->{FH} = $fh;
   $self->{LH} = $type;
   if ($type eq 'L')
   {
      (!$self->{WR}) || die "Writing LAHEY files not supported\n";
      my $buffer;
      (read($self->{FH},$buffer,1)==1) || die "Error reading file header\n";
      (unpack("C",$buffer) == 0xFD) || die "File is not a Lahey file\n";
   }
   return $self;
}

#
#  Rewind file
#
sub rewind
{
   my $self = shift;
   seek($self->{FH},0,SEEK_SET) || return;
   if ($self->{LH} eq 'L')
   {
      my $buffer;
      (read($self->{FH},$buffer,1)==1) || die "Error reading file header\n";
      (unpack("C",$buffer) == 0xFD) || die "File is not a Lahey file\n";
   }
   return 1;
}

#
#  Backspace file
#
sub backspace
{
   my $self = shift;
   my ($n) = @_;
   defined($n) || ($n = 1);
   ($n>0) || die "Invalid backspace count $n\n";
   for (;$n>0;$n--)
   {
      #  Backspace beyond BOF
      (tell($self->{FH})>1) || return;
      if ($self->{LH} eq 'L')
      {
         die "Cannot backspace on Lahey files\n";
      }
      elsif ($self->{LH} eq '0')
      {
         die "Cannot backspace on stream files\n";
      }
      else
      {
         seek($self->{FH},-$self->{LH},SEEK_CUR) || return;
         my $buffer;
         (read($self->{FH},$buffer,$self->{LH})==$self->{LH}) || die "Error reading file header\n";
         my $n = unpack("L",$buffer);
         seek($self->{FH},-2*$self->{LH}-$n,SEEK_CUR) || return;
      }
   }
   return 1;
}

#
#  Close file
#
sub close
{
   my $self = shift;
   close($self->{FH});
}

#
#  Read record head
#
sub head
{
   my $self = shift;
   my $n = shift;
   if ($self->{LH} eq 'L')
   {
      my $buffer;
      (read($self->{FH},$buffer,1)==1) || return;
      my $k = unpack("C",$buffer);
      my $m = ($k&0x03);
      for (my $i=1;$i<=$m;$i++)
      {
         (read($self->{FH},$buffer,1)==1) || die "Error reading record header";
         $k += unpack("C",$buffer)<<(8*$i);
      }
      $n = int($k/4);
   }
   elsif ($self->{LH} ne '0')
   {
      my $buffer;
      (read($self->{FH},$buffer,$self->{LH})==$self->{LH}) || return;
      $n = unpack("L",$buffer);
   }
   return $n;
}

#
#  Read record tail
#
sub tail
{
   my $self = shift;
   my ($n) = @_;
   if ($self->{LH} eq 'L')
   {
      #  Record length including head and tail
      my $k = ($n<<2);
      if (0xFF000000 & $k)
      { $k += 8*4+3; }
      elsif (0xFFFF0000 & $k)
      { $k += 4*4+2; }
      elsif (0xFFFFFF00 & $k)
      { $k += 2*4+1; }
      else
      { $k += 1*4+0; }
      #  Tail length
      my $l=0;
      for (my $m=$k;$m>0;$m = ($m>>8))
      {
         my $buffer;
         (read($self->{FH},$buffer,1)==1) || die "Error reading record tail";
         $l = ($l<<8) + unpack("C",$buffer);
      }
#     printf "n=%d k=%d l=%d l3=%d k3=%d l/4=%d k/4=%d\n" , $n , $k , $l , ($l&3) , ($k&3) , int($l/4) , int($k/4);
#    ($l == $k) || die "Head and tail mismatch\n";
   }
   elsif ($self->{LH} ne '0')
   {
      my $buffer;
      (read($self->{FH},$buffer,$self->{LH})==$self->{LH}) || die "Error reading record tail\n";
      my $l = unpack("L",$buffer);
      ($l == $n) || die "Head and tail mismatch $l != $n\n";
   }
}

#
#  Read record
#
sub read
{
   my $self = shift;
   my $n = $self->head(shift);
   my $buffer;
   ($n) || return;
   my $l = read($self->{FH},$buffer,$n);
   ($self->{LH} eq '0' && $l==0) && return;
   ($l==$n) || die "Error reading file ($l/$n)\n";
   $self->tail($n);
   return $buffer;
}

#
#  Read record and unpack
#
sub readunpack
{
   my $self = shift;
   my $n = $self->head(shift);
   my $fmt = shift;
   my $buffer = $self->read($n);
   defined($buffer) || return;
   return unpack($fmt,$buffer);
}

#
#  Skip record
#
sub skip
{
   my $self = shift;
   my ($N,$n) = @_;
   defined($N) || ($N = 1);
   ($N>0) || return 1;
   for (;$N>0;$N--)
   {
      my $k = $self->head($n);
      ($k) || return;
      seek($self->{FH},$k,SEEK_CUR) || die "Error seeking file\n";
      $self->tail($k);
   }
   return 1;
}

#
#  Write record
#
sub write
{
   my $self = shift;
   my ($buffer) = @_;
   my $fh = $self->{FH};
   if ($self->{LH}==8)
   {
      return print $fh pack("L2a*L2",length($buffer),0,$buffer,length($buffer),0);
   }
   elsif ($self->{LH}==4)
   {
      return print $fh pack("La*L",length($buffer),$buffer,length($buffer));
   }
   elsif ($self->{LH}==0)
   {
      return print $fh $buffer;
   }
   else
   {
      die "Invalid write type $self->{LH}\n";
   }
}

1;
