#!/usr/bin/perl -w
#
#  Geometric Operations
#  Use this module to access and write DBF files
#
#  Copyright (c) 2001  Willem A. Schreuder <willem@prinmath.com>
#  This file is part of the KUDU suite of software.
#
#  The KUDU programs are free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License as
#  published by the Free Software Foundation; either version 2 of the
#  License, or (at your option) any later version.
#
#  The KUDU programs are distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  General Public License for more details.
#
#  You should have received a copy of the GNU General Public License along 
#  with this program; if not, write to the Free Software Foundation, Inc.,
#  51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
#
#  See also www.fsf.org

package Geometry;
use strict;

#  Mask for big and little endian systems
my $oddmask  = pack("L2",substr(pack("s",0x424C),0,1) eq "B" ? ( 0, 1) : ( 1, 0));
my $evenmask = pack("L2",substr(pack("s",0x424C),0,1) eq "B" ? (~0,~1) : (~1,~0));

#
#  Return an even double
#
sub evenreal
{
   return unpack("d",pack("d",$_[0])&"$evenmask");
}

#
#  Return an odd double
#
sub oddreal
{
   return unpack("d",pack("d",$_[0])|"$oddmask");
}

#
#  Check if two line segments intersect
#
sub IntersectF
{
   my ($x1,$y1,$x2,$y2,$x3,$y3,$x4,$y4) = @_;
   my $tiny = 1e-15;
   my $zero = $tiny;
   my $one  = 1-$tiny;

   my $xmin12 = ($x1<$x2) ? $x1 : $x2;
   my $xmax12 = ($x1>$x2) ? $x1 : $x2;
   my $ymin12 = ($y1<$y2) ? $y1 : $y2;
   my $ymax12 = ($y1>$y2) ? $y1 : $y2;
   my $xmin34 = ($x3<$x4) ? $x3 : $x4;
   my $xmax34 = ($x3>$x4) ? $x3 : $x4;
   my $ymin34 = ($y3<$y4) ? $y3 : $y4;
   my $ymax34 = ($y3>$y4) ? $y3 : $y4;

   #  No intersection
   my $f;
   if ($xmax12<$xmin34 || $xmin12>$xmax34 ||
       $ymax12<$ymin34 || $ymin12>$ymax34)
   {
      $f = undef;
   }
   else
   {
      my $x12 = $x1-$x2;
      my $y12 = $y1-$y2;
      my $x43 = $x4-$x3;
      my $y43 = $y4-$y3;
      my $x42 = $x4-$x2;
      my $y42 = $y4-$y2;
      my $x31 = $x3-$x1;
      my $y31 = $y3-$y1;
      my $det = $x12*$y43-$x43*$y12;
      my $gam = (abs($x12)+abs($x43)+abs($y12)+abs($y43))/4;
      #  Segments on crossing lines
      if (abs($det)>$tiny*$gam)
      {
         my $f1 = ($y43*$x42-$x43*$y42)/$det;
         my $f2 = ($x12*$y42-$y12*$x42)/$det;
         $f = ($zero<$f1 && $f1<$one && $zero<$f2 && $f2<$one) ? 1-$f1 : undef;
      }
      #  Segments do not cross
      else
      {
         $f = undef;
      }
   }

   return $f;
}

#
#  Check if two line segments intersect
#
sub intersect
{
   my ($x1,$y1,$x2,$y2,$x3,$y3,$x4,$y4) = @_;
   my $f = IntersectF(@_) || return;
   my $x = (1-$f)*$x1 + $f*$x2;
   my $y = (1-$f)*$y1 + $f*$y2;
   return ($x,$y);
}

#
#  Check if point is inside triangle
#
sub intri
{
   my $tiny = 1e-15;
   my ($x,$y,$x1,$y1,$x2,$y2,$x3,$y3) = @_;

   my $det = ($x1-$x2)*($y-$y3)-($x-$x3)*($y1-$y2);
   if (abs($det)>$tiny)
   {
      my $f1 = (($x-$x2)*($y-$y3)-($x-$x3)*($y-$y2))/$det;
      my $f2 = (($x1-$x2)*($y-$y2)-($y1-$y2)*($x-$x2))/$det;
      return (0<=$f1 && $f1<=1 && $f2<=0);
   }
   else
   {
      return 0;
   }
}

#
#  Find corner
#
sub dirmax
{
   my ($x,$y,$dir) = @_;
   $dir = uc($dir);
   my $k = 0;
   if ($dir eq 'NW')
   {
      for (my $i=1;$i<@$x;$i++)
      {
         ($y->[$i]-$x->[$i] > $y->[$k]-$x->[$k]) && ($k = $i);
      }
   }
   elsif ($dir eq 'SE')
   {
      for (my $i=1;$i<@$x;$i++)
      {
         ($y->[$i]-$x->[$i] < $y->[$k]-$x->[$k]) && ($k = $i);
      }
   }
   elsif ($dir eq 'NE')
   {
      for (my $i=1;$i<@$x;$i++)
      {
         ($y->[$i]+$x->[$i] > $y->[$k]+$x->[$k]) && ($k = $i);
      }
   }
   elsif ($dir eq 'SW')
   {
      for (my $i=1;$i<@$x;$i++)
      {
         ($y->[$i]+$x->[$i] < $y->[$k]+$x->[$k]) && ($k = $i);
      }
   }
   elsif ($dir eq 'N')
   {
      for (my $i=1;$i<@$x;$i++)
      {
         ($y->[$i] > $y->[$k]) && ($k = $i);
      }
   }
   elsif ($dir eq 'S')
   {
      for (my $i=1;$i<@$x;$i++)
      {
         ($y->[$i] < $y->[$k]) && ($k = $i);
      }
   }
   elsif ($dir eq 'E')
   {
      for (my $i=1;$i<@$x;$i++)
      {
         ($x->[$i] > $x->[$k]) && ($k = $i);
      }
   }
   elsif ($dir eq 'W')
   {
      for (my $i=1;$i<@$x;$i++)
      {
         ($x->[$i] < $x->[$k]) && ($k = $i);
      }
   }
   else
   {
      die "Unknown dirmax direction $dir\n";
   }
   return ($x->[$k],$y->[$k]);
}

#
#  Return points that distance d1,d2 from (x1,y1) and (x2,y2)
#
sub xydist
{
   (@_==6) || die "Wrong number of arguments to xydist\n";
   my ($x1,$y1,$d1,$x2,$y2,$d2) = @_;
   my $d = sqrt(($x2-$x1)**2+($y2-$y1)**2);
   #  Angle
   my $cos = ($d**2+$d1**2-$d2**2) / (2*$d*$d1);
   #  No solution
   (abs($cos)>1) && return ($d,$d1,$d2);
   my $sin = sqrt(1-$cos**2);
   #  Vector length d1 from 1 to 2
   my $ex = $d1*($x2-$x1)/$d;
   my $ey = $d1*($y2-$y1)/$d;
   #  Rotate left
   my $xl = +$ex*$cos + $ey*$sin +$x1;
   my $yl = -$ex*$sin + $ey*$cos +$y1;
   #  Rotate right
   my $xr = +$ex*$cos - $ey*$sin +$x1;
   my $yr = +$ex*$sin + $ey*$cos +$y1;
   return ($xl,$yl,$xr,$yr);
}

#
#  Calculate Liang-Barsky clipping t
#    den is N(P0-P1)
#    num is N(P0-Pe)
#  For a window parallel to x and y axes:
#    den is +/-dx or +/-dy
#    num is the perpendicular distance to P0
#/
sub LBclipT
{
   my ($den,$num,$t0,$t1) = @_;
   #  Potential entry
   if ($den>0)
   {
      my $t = $num/$den;
      ($t>$$t1) && return;      #  Crossover => reject
      ($t>$$t0) && ($$t0 = $t); #  New entry
   }
   #  Potential exit
   elsif ($den<0)
   {
      my $t = $num/$den;
      ($t<$$t0) && return;      #  Crossover => reject
      ($t<$$t1) && ($$t1 = $t); #  New exit
   }
   #  On outside edge => reject
   elsif ($num>0)
   {
      return;
   }
   #  Return Accept
   return 1;
}

#
#  Clip line
#
sub ClipLine
{
   my ($x0,$y0,$x1,$y1,$Xmin,$Xmax,$Ymin,$Ymax) = @_;
   my $dx = $x1-$x0;
   my $dy = $y1-$y0;
   my $t0 = 0;
   my $t1 = 1;
   #  Check for degenerate lines
   if ($dx==0 && $dy==0)
   {
      return ($Xmin<=$x0 && $x0<=$Xmax && $Ymin<=$y0 && $y0<=$Ymax) ? ($x0,$y0,$x1,$y1) : ();
   }
   #  Calculate t for all four edges
   elsif (LBclipT(+$dx,$Xmin-$x0,\$t0,\$t1) &&  # Left
          LBclipT(-$dx,$x0-$Xmax,\$t0,\$t1) &&  # Right
          LBclipT(+$dy,$Ymin-$y0,\$t0,\$t1) &&  # Bottom
          LBclipT(-$dy,$y0-$Ymax,\$t0,\$t1))    # Top
   {
      if ($t1<1)
      {
         $x1 = $x0 + $t1*$dx;
         $y1 = $y0 + $t1*$dy;
      }
      if ($t0>0)
      {
         $x0 += $t0*$dx;
         $y0 += $t0*$dy;
      }
      return ($x0,$y0,$x1,$y1);
   }
   #  Complete miss
   else
   {
      return;
   }
}

#
#  Clip line with interpolation
#
sub ClipLineZ
{
   my ($x0,$y0,$z0,$x1,$y1,$z1,$Xmin,$Xmax,$Ymin,$Ymax) = @_;
   my $dx = $x1-$x0;
   my $dy = $y1-$y0;
   my $dz = $z1-$z0;
   my $t0 = 0;
   my $t1 = 1;
   #  Check for degenerate lines
   if ($dx==0 && $dy==0)
   {
      return ($Xmin<=$x0 && $x0<=$Xmax && $Ymin<=$y0 && $y0<=$Ymax) ? ($x0,$y0,$z0,$x1,$y1,$z1) : ();
   }
   #  Calculate t for all four edges
   elsif (LBclipT(+$dx,$Xmin-$x0,\$t0,\$t1) &&  # Left
          LBclipT(-$dx,$x0-$Xmax,\$t0,\$t1) &&  # Right
          LBclipT(+$dy,$Ymin-$y0,\$t0,\$t1) &&  # Bottom
          LBclipT(-$dy,$y0-$Ymax,\$t0,\$t1))    # Top
   {
      if ($t1<1)
      {
         $x1 = $x0 + $t1*$dx;
         $y1 = $y0 + $t1*$dy;
         $z1 = $z0 + $t1*$dz;
      }
      if ($t0>0)
      {
         $x0 += $t0*$dx;
         $y0 += $t0*$dy;
         $z0 += $t0*$dz;
      }
      return ($x0,$y0,$z0,$x1,$y1,$z1);
   }
   #  Complete miss
   else
   {
      return;
   }
}

#
#  Clip to plane
#
sub ClipPlane
{
   my ($X,$Y,$Xmm,$mm) = @_;
   my (@X,@Y);
   for (my $i=0;$i<@$X;$i++)
   {
      my (@x,@y);
      for (my $j=0;$j<@{$X->[$i]};$j++)
      {
         my $j0 = ($j==0) ? @{$X->[$i]}-1 : $j-1;
         my $i0 = ($mm*($Xmm-$X->[$i][$j0])>=0);
         my $i1 = ($mm*($Xmm-$X->[$i][$j])>=0);
         #  Both in
         if ($i0 && $i1)
         {
            push @x , $X->[$i][$j];
            push @y , $Y->[$i][$j];
         }
         #  Cross in or out
         elsif ($i0 || $i1)
         {
            my $f = ($Xmm-$X->[$i][$j0])/($X->[$i][$j]-$X->[$i][$j0]);
            push @x , $Xmm;
            push @y , (1-$f)*$Y->[$i][$j0] + $f*$Y->[$i][$j];
            #  End point is in
            if ($i1)
            {
               push @x , $X->[$i][$j];
               push @y , $Y->[$i][$j];
            }
         }
      }
      if (@x>0)
      {
         push @X , \@x;
         push @Y , \@y;
      }
   }
   return (\@X,\@Y);
}

#
#  Clip polygon
#
sub ClipPoly
{
   my ($X,$Y,$Xmin,$Xmax,$Ymin,$Ymax) = @_;
   ($X,$Y) = ClipPlane($X,$Y,$Xmin,-1);
   ($X,$Y) = ClipPlane($X,$Y,$Xmax,+1);
   ($Y,$X) = ClipPlane($Y,$X,$Ymin,-1);
   ($Y,$X) = ClipPlane($Y,$X,$Ymax,+1);
   return ($X,$Y);
}

#
#  Calculate polygon area
#
sub PolyArea
{
   my ($X,$Y) = @_;
   my $A=0;
   for (my $i=0;$i<@$X;$i++)
   {
      my $n = @{$X->[$i]}-1;
      my $a;
      for (my $j=0;$j<$n;$j++)
      {
         $a += ($X->[$i][$j+1]-$X->[$i][$j]) * ($Y->[$i][$n-$j]-$Y->[$i][$j])
             - ($Y->[$i][$j+1]-$Y->[$i][$j]) * ($X->[$i][$n-$j]-$X->[$i][$j]);
      }
      $A += ($i==0) ? abs($a) : -abs($a);
   }
   return $A/2;
}

1;
