#!/usr/bin/perl -w
#
#  PM Utility Programs
#  This package provides some convenience routines that can be used
#  to access files that are written in DIAMOND format.
#
#  Copyright (c) 2001  Willem A. Schreuder <willem@prinmath.com>
#  This file is part of the KUDU suite of software.
#
#  The KUDU programs are free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License as
#  published by the Free Software Foundation; either version 2 of the
#  License, or (at your option) any later version.
#
#  The KUDU programs are distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  General Public License for more details.
#
#  You should have received a copy of the GNU General Public License along 
#  with this program; if not, write to the Free Software Foundation, Inc.,
#  51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
#
#  See also www.fsf.org

package PM;
use strict;

#
#  grflib root
#
sub grflib
{
   my $grflib = '/pm/grflib/'.join('',@_);
   ($0 =~ /^([A-Z]:)/) && ($grflib = "$1$grflib");
   return $grflib;
}

#
#  Split honoring quotes and commas
#
sub Split
{
   my ($line) = @_;
   my @line;
   my $status = -1;
   my $field = undef;
   for (my $i=0;$i<length($line);$i++)
   {
      my $ch = substr($line,$i,1);
      (ord($ch)==146) && ($ch = "'");  #  WordPerfect insanity
      #  A quote
      if ($ch eq '"')
      {
         if ($status>0)
         {
            #  Check if repeated
            if ($i+1<length($line) && substr($line,$i+1,1) eq '"')
            {
               $field .= '"';
               $i++;
            }
            #  End of quote
            else
            {
               push @line , defined($field) ? $field : "";
               $field = undef;
               $status = 0;
            }
         }
         else
         {
            $status = 1;
         }
      }
      #  All's fair in quotes
      elsif ($status>0)
      {
         $field .= $ch;
      }
      #  Unquoted whitespace
      elsif ($ch =~ '\s')
      {
         if (defined($field))
         {
            push @line , $field;
            $field = undef;
         }
      }
      #  Unquoted comma
      elsif ($ch eq ',')
      {
         if ($status < 0 || defined($field))
         {
            push @line , defined($field) ? $field : "";
            $field = undef;
         }
         $status = -1;
      }
      #  Gotta be something
      else
      {
         $field .= $ch;
         $status = 0;
      }
   }
   ($status<=0) || die "Unclosed quote: $line";
   if (defined($field))
   {
      push @line , $field;
   }
   return @line;
}

#
#  Find entry in an array
#
sub Index
{
   my ($txt,@txt) = @_;
   for (my $k=0;$k<@txt;$k++)
   {
      ($txt eq $txt[$k]) && return $k;
   }
   return;
}

#
#  Extract DIAMOND real
#
sub Real
{
   my @real = @_;
   for (my $k=0;$k<@real;$k++)
   {
      my $real = uc($real[$k]);
      $real =~ s/^ *//;
      $real =~ s/ *$//;
      if ($real =~ /NAN/)
      {
         $real[$k] = "NaN";
      }
      elsif ($real =~ ":")
      {
         my @dms = split(":",$real);
         my $div = 1;
         my $dms = abs($dms[0]);
         for (my $k=1;$k<@dms;$k++)
         {
            $div *= 60;
            $dms += $dms[$k]/$div;
         }
         $real[$k] = (substr($real,0,1) eq "-") ? -$dms : $dms;
      }
   }
   return (wantarray() ? @real : $real[0]);
}

#
#  Get substring with trim
#
sub substrim
{
   my ($line,$off,$len) = @_;
   my $str = defined($len) ? substr($line,$off,$len) : substr($line,$off);
   if (defined($str))
   {
      $str =~ s/^  *//;
      $str =~ s/  *$//;
   }
   return $str;
}

#
#  Trim string
#
sub trim
{
   my ($str) = @_;
   if (defined($str))
   {
      $str =~ s/^  *//;
      $str =~ s/  *$//;
   }
   return $str;
}

#
#  Convert Real to DMS
#
sub DMS
{
   my @real = @_;
   for (my $k=0;$k<@real;$k++)
   {
      my $real = uc($real[$k]);
      $real =~ s/^ *//;
      $real =~ s/ *$//;
      if ($real =~ /NAN/)
      {
         $real[$k] = 'NaN';
      }
      else
      {
         my $deg = int($real);
         $real = 60*abs($real-$deg);
         my $min = int($real);
         $real = 60*($real-$min);
         my $sec = $real;
         if ($sec>59.99)
         {
            $sec = 0;
            $min++;
         }
         if ($min>59)
         {
            $min -= 60;
            $deg += ($deg<0) ? -1 : +1;
         }
         $real[$k] = sprintf '%d:%.2d:%05.2f' , $deg,$min,$sec;
      }
   }
   return (wantarray() ? @real : $real[0]);
}

1;
