#!/usr/bin/perl -w
#
#  Transform Object
#  Use this module to perform coordinate transformations
#
#  Copyright (c) 2001  Willem A. Schreuder <willem@prinmath.com>
#  This file is part of the KUDU suite of software.
#
#  The KUDU programs are free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License as
#  published by the Free Software Foundation; either version 2 of the
#  License, or (at your option) any later version.
#
#  The KUDU programs are distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  General Public License for more details.
#
#  You should have received a copy of the GNU General Public License along 
#  with this program; if not, write to the Free Software Foundation, Inc.,
#  51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
#
#  See also www.fsf.org

package Transform;

use strict;
use FileHandle;
use Math::Trig;
use PM;
use IO::Seekable;

#
#  Constants
#
my $pi  = 3.141592653589793238462643;
my $d2r = $pi/180;
my $r2d = 180/$pi;
my $tol = 1e-6;
my $spcs = PM::grflib('data/spcs.dat');
my %m2x  = (M => 1 , SFT => 39.37/12 , IFT => 1/0.3048);
my %x2m  = (M => 1 , SFT => 12/39.37 , IFT => 0.3048);
my $MI = (unpack("s",pack("v",1))!=1) ? 'M' : 'I';

#
#  Binary read utilities
sub GetI16
{
   my ($fd,$mi) = @_;
   my $buf;
   (read($fd,$buf,2)==2) || die "Cannot GetI16\n";
   return unpack('S',$mi ne $MI ? (reverse $buf) : $buf);
}
sub GetI32
{
   my ($fd,$mi) = @_;
   my $buf;
   (read($fd,$buf,4)==4) || die "Cannot GetI32\n";
   return unpack('L',$mi ne $MI ? (reverse $buf) : $buf);
}
sub GetR32
{
   my ($fd,$mi) = @_;
   my $buf;
   (read($fd,$buf,4)==4) || die "Cannot GetR32\n";
   return unpack('f',$mi ne $MI ? (reverse $buf) : $buf);
}

#
#  Create a Transform object
#
sub new
{
   my $class = shift;
   my $self = {};
   bless($self,$class);

   #  Hash
   if (@_>1)
   {
      push @{$self->{XFM}} , Transform->simple(@_);
   }
   #  Parse text string for transform
   else 
   {
      push @{$self->{XFM}} , ParseString(@_);
   }
   return $self;
}

#
#  Parse string and return array of simple objects
#
sub ParseString
{
   my @xfm;
   foreach my $str (split("\\|",$_[0]))
   {
      my @str = split(' ',$str);
      my ($inv,%args,%spcs);
      while (@str>0)
      {
         my $arg = uc shift @str;
         if ($arg eq 'OFFSET')
         {
            $args{X0} = shift @str;
            $args{Y0} = shift @str;
         }
         elsif ($arg eq 'PIVOT')
         {
            $args{Xp} = shift @str;
            $args{Yp} = shift @str;
         }
         elsif ($arg eq 'ROTATE')
         {
            $args{ROTATE} = shift @str;
         }
         elsif ($arg eq 'SCALE')
         {
            $args{SCALE} = shift @str;
         }
         elsif ($arg eq 'INV')
         {
            $inv = 1;
         }
         elsif ($arg eq 'PROJECT')
         {
            my $projection = shift @str;
            ($projection =~ s/^(.*):(.*)$/$1/) && ($args{ELLIPSE} = $2);
            
            if ($projection eq 'LCC' || $projection eq 'AEC')
            {
               (@str>=2) || die "Missing $projection parameters $str\n";
               $args{PROJECTION} = $projection;
               $args{LON0} = shift @str;
               $args{LAT0} = shift @str;
               (@str>0 && $str[0] !~ /[A-Z]/i) && ($args{LAT1} = shift @str);
               (@str>0 && $str[0] !~ /[A-Z]/i) && ($args{LAT2} = shift @str);
            }
            elsif ($projection eq 'APC')
            {
               (@str>=2) || die "Missing $projection parameters $str\n";
               $args{PROJECTION} = $projection;
               $args{LON0} = shift @str;
               $args{LAT0} = shift @str;
            }
            elsif ($projection eq 'TM')
            {
               (@str>=1) || die "Missing $projection parameters $str\n";
               $args{PROJECTION} = $projection;
               $args{LON0} = shift @str;
               (@str>0 && $str[0] !~ /[A-Z]/i) && ($args{LAT0} = shift @str);
               (@str>0 && $str[0] !~ /[A-Z]/i) && ($args{K0}   = shift @str);
            }
            elsif ($projection eq 'MER')
            {
               (@str>=1) || die "Missing $projection parameters $str\n";
               $args{PROJECTION} = $projection;
               $args{LON0} = shift @str;
            }
            elsif ($projection eq 'NAD83')
            {
               $args{PROJECTION} = $projection;
            }
            elsif (%spcs = GetSPCS($projection))
            {
            }
            else
            {
               die "Unknown projection $str\n";
            }
         }
         else
         {
            die "Unknown parameter $arg\n";
         }
      }
      (%spcs && %args) && die "Parameters invalid with SPCS $str\n";
      %spcs && (%args = %spcs);
      $inv && ($args{INV} = 1);
      push @xfm , Transform->simple(%args);
   }
   return @xfm;
}

#
#  Set up a Transform object
#
sub simple
{
   my $class = shift;
   my %args = (ELLIPSE    => '1866' ,
               PROJECTION => '' ,
               LON0       => 0 ,
               LAT0       => 0 ,
               LAT1       => undef ,
               LAT2       => undef ,
               K0         => undef ,
               X0         => 0 ,
               Y0         => 0 ,
               Xp         => 0 ,
               Yp         => 0 ,
               ROTATE     => 0 ,
               SCALE      => 1 ,
               INV        => 0 ,
               @_);
   my $self = \%args;
   bless($self,$class);

   my $projection = $self->{PROJECTION};

   #
   #  Set the ellipse if a real projection
   #
   $projection && ($self->ellipse());
  
   if (!$projection)
   {
      $self->{FORWARD} = \&NULL;
      $self->{INVERSE} = \&NULL;
      $self->{SFACTOR} = \&UNIT;
   }
   elsif ($projection eq 'TM')
   {
      defined($self->{K0}) || ($self->{K0} = 0.9996);
      $self->TMset();
      $self->{FORWARD} = \&TMforward;
      $self->{INVERSE} = \&TMinverse;
      $self->{SFACTOR} = \&TMscale;
   }
   elsif ($projection eq 'LCC')
   {
      defined($self->{LAT1}) || ($self->{LAT1} = 33.0);
      defined($self->{LAT2}) || ($self->{LAT2} = 45.0);
      $self->LCCset();
      $self->{FORWARD} = \&LCCforward;
      $self->{INVERSE} = \&LCCinverse;
      $self->{SFACTOR} = \&LCCscale;
   }
   elsif ($projection eq 'AEC')
   {
      defined($self->{LAT1}) || ($self->{LAT1} = 29.5);
      defined($self->{LAT2}) || ($self->{LAT2} = 45.5);
      $self->AECset();
      $self->{FORWARD} = \&AECforward;
      $self->{INVERSE} = \&AECinverse;
      $self->{SFACTOR} = \&AECscale;
   }
   elsif ($projection eq 'NAD83')
   {
      $self->NAD83set();
      $self->{FORWARD} = \&NAD83forward;
      $self->{INVERSE} = \&NAD83inverse;
      $self->{SFACTOR} = \&NAD83scale;
   }
   else
   {
      die "Unknown projection $projection\n";
   }

   my $th = $d2r*$self->{ROTATE};
   my $r  = $self->{SCALE};
   my $costh = cos($th);
   my $sinth = sin($th);
   $self->{rcosth} = $r*$costh;
   $self->{rsinth} = $r*$sinth;
   $self->{costhr} = $costh/$r;
   $self->{sinthr} = $sinth/$r;
   my $X0 = $self->{X0};
   my $Y0 = $self->{Y0};
   my $Xp = $self->{Xp};
   my $Yp = $self->{Yp};
   $self->{xoff} = $X0 + $Xp*(1-$r*$costh) + $Yp*$r*$sinth;
   $self->{yoff} = $Y0 + $Yp*(1-$r*$costh) - $Xp*$r*$sinth;

   return $self;
}

#
#  Forward transformation
#
sub forward
{
   my $self = shift;
   my ($x,$y) = @_;
   return $self->xfm($x,$y,0);
}

#
#  Inverse transformation
#
sub inverse
{
   my $self = shift;
   my ($x,$y) = @_;
   return $self->xfm($x,$y,1);
}

#
#  Transformation
#
sub xfm
{
   my $self = shift;
   my ($x,$y,$inv) = @_;
   foreach my $xfm ($inv ? reverse @{$self->{XFM}} : @{$self->{XFM}})
   {
      #  Inverse (x,y -> lon,lat)
      if ($inv xor $xfm->{INV})
      {
         #  Unpack
         my $projection = $xfm->{PROJECTION};
         my $x0         = $xfm->{xoff};
         my $y0         = $xfm->{yoff};
         my $costhr     = $xfm->{costhr};
         my $sinthr     = $xfm->{sinthr};
         my $inverse    = $xfm->{INVERSE};
         #  Planar transformation
         my $x1 =  $costhr*($x-$x0) + $sinthr*($y-$y0);
         my $y1 = -$sinthr*($x-$x0) + $costhr*($y-$y0);
         #  Projection
         ($x,$y) = $xfm->$inverse($x1,$y1);
         }
      #  Forward (lon,lat -> x,y)
      else
      {
         #  Unpack
         my $projection = $xfm->{PROJECTION};
         my $x0         = $xfm->{xoff};
         my $y0         = $xfm->{yoff};
         my $rcosth     = $xfm->{rcosth};
         my $rsinth     = $xfm->{rsinth};
         my $forward    = $xfm->{FORWARD};
         #  Projection
         my ($x1,$y1) = $xfm->$forward($x,$y);
         #  Planar transformation
         $x = $rcosth*$x1 - $rsinth*$y1 + $x0;
         $y = $rsinth*$x1 + $rcosth*$y1 + $y0;
      }
   }
   return ($x,$y);
}

#
#  Scale adjustment (lon,lat -> scale)
#
sub scale
{
   my $self = shift;
   my ($lon,$lat) = @_;
   my $scale = 1;
   foreach my $xfm (@{$self->{XFM}})
   {
      my $sfactor = $xfm->{SFACTOR};
      my $f = $xfm->$sfactor($lon,$lat);
      $scale *= $xfm->{INV} ? 1/$f : $f;
   }
   return $scale;
}

#
#  NULL transformation
#
sub NULL
{
   my $self = shift;
   return @_;
}

#
#  UNIT transformation
#
sub UNIT
{
   return 1;
}

#
#  Set the ellipse and related parameters
#
sub ellipse
{
   my $self = shift;

   my ($a,$b);
   if ($self->{ELLIPSE} eq '1866')
   {
      $a = 6378206.4;
      $b = 6356583.8;
   }
   elsif ($self->{ELLIPSE} eq 'GRS80' || $self->{ELLIPSE} eq 'WGS84')
   {
      $a = 6378137.0;
      $b = 6356752.3;
   }
   elsif ($self->{ELLIPSE} eq 'SPH84')
   {
      $a = 6378137.0;
      $b = 6378137.0;
   }
   elsif ($self->{ELLIPSE} eq 'WGS72')
   {
      $a = 6378135.0;
      $b = 6356750.5;
   }
   elsif ($self->{ELLIPSE} eq 'MICH')
   {
      $a = 1.0000382*6378206.4;
      $b = 1.0000382*6356583.8;
   }
   elsif ($self->{ELLIPSE} eq 'INTL')
   {
      $a = 6378388;
      $b = 296/297*$a;
   }
   else
   {
      die "Unknown ellipsoid $self->{ELLIPSE}\n";
   }

   #  Eccentricity, powers and functions
   my $e2  = 1-($b*$b)/($a*$a);
   my $e   = sqrt($e2);
   my $e4  = $e2*$e2;
   my $e6  = $e4*$e2;
   my $ee2 = $e2/(1-$e2);

   #  Store parameters
   $self->{a}   = $a;
   $self->{b}   = $b;
   $self->{e}   = $e;
   $self->{e2}  = $e2;
   $self->{e4}  = $e4;
   $self->{e6}  = $e6;
   $self->{ee2} = $ee2;
}

#
#  Special transformations
#
sub UTM
{
   my $class = shift;
   my ($zone,$nad,$scale) = @_;
   defined($nad) || ($nad = 'NAD27');
   defined($scale) || ($scale = 1);
   ($zone == int($zone) && 1<=$zone && $zone<=60) || die "Invalid UTM zone $zone";
   ($nad eq 'NAD27') || ($nad eq 'NAD83') || die "Invalid UTM datum $nad\n";
   ($scale>0) || die "Invalid UTM scale $scale\n";
   my $elps = ($nad eq 'NAD83') ? 'GRS80' : '1866';
   return Transform->new(PROJECTION => 'TM' , LON0 => 6*$zone-183 , K0 => 0.9996 , X0 => 500000*$scale , ELLIPSE => $elps , SCALE => $scale);
}

sub NAD83
{
   my $class = shift;
   return Transform->new(PROJECTION => 'NAD83');
}

#
#  Read SPCS file for this type
sub GetSPCS
{
   my ($type) = @_;
   open(SPCS , "<$spcs") || die "Cannot open file $spcs";
   <SPCS>;
   my ($code,$projection,$lon0,$lat0,$p1,$p2,$ellipse,$dx,$dy,$Uoff,$Unit,$datum,$theta,$xp,$yp,$name);
   while (my $line = <SPCS>)
   {
      chomp $line;
      if (substr($line,0,1) ne '#')
      {
         ($code,$projection,$lon0,$lat0,$p1,$p2,$ellipse,$dx,$dy,$Uoff,$Unit,$datum,$theta,$xp,$yp,$name) = PM::Split($line);
         last if ($code eq 'END' || $code eq $type);
      }
   }
   close(SPCS);
   ($code eq $type) || return;

   #  Output units
   $Unit = uc($Unit);
   $Unit = 'SFT' if ($Unit eq 'FT');
   (exists $m2x{$Unit}) || die "Invalid SPCS units $Unit";
   my $scale = $m2x{$Unit};

   #  Offset units
   $Uoff = uc($Uoff);
   $Uoff = 'SFT' if ($Uoff eq 'FT');
   (exists $m2x{$Uoff}) || die "Invalid SPCS units $Uoff";
   if ($Uoff ne $Unit)
   {
      $dx *= $x2m{$Uoff}*$m2x{$Unit};
      $dy *= $x2m{$Uoff}*$m2x{$Unit};
   }

   #  Rotation
   my $costh = cos($d2r*$theta);
   my $sinth = sin($d2r*$theta);
   my $x0 = $dx*$costh - ($dy-$yp)*$sinth + $xp*(1-$costh);
   my $y0 = $dy*$costh + ($dx-$xp)*$sinth + $yp*(1-$costh);
 
   ($lon0,$lat0,$p1,$p2) = PM::Real($lon0,$lat0,$p1,$p2);

   #  Set parameters
   if ($projection eq 'TM')
   {
      ($p1>1) && ($p1 = 1-1/$p1);
      return (PROJECTION=>'TM' , LON0=>$lon0 , LAT0=>$lat0 , K0=>$p1 , X0=>$x0 , Y0=>$y0 , ELLIPSE=>$ellipse , SCALE=>$scale , ROTATE=>$theta);
   }
   elsif ($projection eq "LCC")
   {
      return (PROJECTION=>'LCC' , LON0=>$lon0 , LAT0=>$lat0 , LAT1=>$p1 , LAT2=>$p2 , X0=>$x0 , Y0=>$y0 , ELLIPSE=>$ellipse , SCALE=>$scale , ROTATE=>$theta);
   }
   elsif ($projection eq "AEC")
   {
      return (PROJECTION=>'AEC' , LON0=>$lon0 , LAT0=>$lat0 , LAT1=>$p1 , LAT2=>$p2 , X0=>$x0 , Y0=>$y0 , ELLIPSE=>$ellipse , SCALE=>$scale , ROTATE=>$theta);
   }
   else
   {
      die "Invalid SPCS projection $projection\n";
   }
}

sub SPCS
{
   my $class = shift;
   my ($type) = @_;
   my $self;
   my %args = GetSPCS($type);
   %args || die "Unknown SPCS type $type\n";
   return Transform->new(%args);
}

#
#  Transverse Mercator Projection:  Set
#
sub TMset
{
   my $self = shift;

   my $lat0 = $self->{LAT0};
   my $a    = $self->{a};
   my $e2   = $self->{e2};
   my $e4   = $self->{e4};
   my $e6   = $self->{e6};

   my $fi = $d2r*$lat0;
   my $E1 = (1-sqrt(1-$e2))/(1+sqrt(1-$e2));
   my $E2 = $E1*$E1;
   my $E3 = $E2*$E1;
   my $E4 = $E3*$E1;
   my $ae0 = $a*(1-$e2/4-3*$e4/64-5*$e6/256);
   my $ae2 = $a*(3*$e2/8+3*$e4/32+45*$e6/1024);
   my $ae4 = $a*(15*$e4/256+45*$e6/1024);
   my $ae6 = $a*(35*$e6/3072);
   my $me2 = (3*$E1/2-27*$E3/32);
   my $me4 = (21*$E2/16-55*$E4/32);
   my $me6 = 151*$E3/96;
   my $me8 = 1097*$E4/512;
   my $M0  = $ae0*$fi-$ae2*sin(2*$fi)+$ae4*sin(4*$fi)-$ae6*sin(6*$fi);

   $self->{ae0} = $ae0;
   $self->{ae2} = $ae2;
   $self->{ae4} = $ae4;
   $self->{ae6} = $ae6;
   $self->{me2} = $me2;
   $self->{me4} = $me4;
   $self->{me6} = $me6;
   $self->{me8} = $me8;
   $self->{M0}  = $M0;
}

#
#  Transverse Mercator Projection:  forward
#
sub TMforward
{
   my $self = shift;
   my ($lon,$lat) = @_;

   my $a    = $self->{a};
   my $k0   = $self->{K0};
   my $M0   = $self->{M0};
   my $lon0 = $self->{LON0};
   my $e2   = $self->{e2};
   my $ee2  = $self->{ee2};
   my $ae0  = $self->{ae0};
   my $ae2  = $self->{ae2};
   my $ae4  = $self->{ae4};
   my $ae6  = $self->{ae6};

   my $fi    = $d2r*$lat;
   my $sinfi = sin($fi);
   my $cosfi = cos($fi);
   my $tanfi = $sinfi/$cosfi;
   my $T     = $tanfi*$tanfi;
   my $T2    = $T*$T;
   my $N     = $a/sqrt(1-$e2*$sinfi*$sinfi);
   my $C     = $ee2*$cosfi*$cosfi;
   my $C2    = $C*$C;
   my $M     = $ae0*$fi-$ae2*sin(2*$fi)+$ae4*sin(4*$fi)-$ae6*sin(6*$fi);
   my $A     = $d2r*($lon-$lon0)*$cosfi;
   my $A2    = $A*$A;
   my $A3    = $A*$A2;
   my $A4    = $A*$A3;
   my $A5    = $A*$A4;
   my $A6    = $A*$A5;

   my $x = $N*($A + (1-$T+$C)*$A3/6 + (5-18*$T+$T2+72*$C-58*$ee2)*$A5/120);
   my $y = $M-$M0 + $N*$tanfi*($A2/2+(5-$T+9*$C+4*$C2)*$A4/24+(61-58*$T+$T2+600*$C-330*$ee2)*$A6/720);
   return ($k0*$x,$k0*$y);
}

#
#  Transverse Mercator Projection:  inverse
#
sub TMinverse
{
   my $self = shift;
   my ($x,$y) = @_;

   my $a    = $self->{a};
   my $k0   = $self->{K0};
   my $M0   = $self->{M0};
   my $lon0 = $self->{LON0};
   my $e2   = $self->{e2};
   my $e4   = $self->{e4};
   my $e6   = $self->{e6};
   my $ee2  = $self->{ee2};
   my $me2  = $self->{me2};
   my $me4  = $self->{me4};
   my $me6  = $self->{me6};
   my $me8  = $self->{me8};

   my $M     = $M0 + $y/$k0;
   my $mu    = $M/$a/(1-$e2/4-3*$e4/64-5*$e6/256);
   my $fi    = $mu + $me2*sin(2*$mu) + $me4*sin(4*$mu) + $me6*sin(6*$mu) + $me8*sin(8*$mu);
   my $sinfi = sin($fi);
   my $cosfi = cos($fi);
   my $tanfi = $sinfi/$cosfi;
   my $T1    = $tanfi*$tanfi;
   my $T2    = $T1*$T1;
   my $C1    = $ee2*$cosfi*$cosfi;
   my $C2    = $C1*$C1;
   my $q2    = 1-$e2*$sinfi*$sinfi;
   my $q     = sqrt($q2);
   my $N1    = $a/$q;
   my $R1    = $N1*(1-$e2)/$q2;
   my $D     = $x/$N1/$k0;
   my $D2    = $D*$D;
   my $D3    = $D*$D2;
   my $D4    = $D*$D3;
   my $D5    = $D*$D4;
   my $D6    = $D*$D5;

   my $lat = $r2d*($fi - $N1*$tanfi/$R1*($D2/2 - (5+3*$T1+10*$C1-4*$C2-9*$ee2)*$D4/24 + (61+90*$T1+298*$C1+45*$T2-252*$ee2-3*$C2)*$D6/720));
   my $lon = $lon0 + $r2d/$cosfi*($D - (1+2*$T1+$C1)*$D3/6 + (5-2*$C1+28*$T1-3*$C2+8*$ee2+24*$T2)*$D5/120);
   return ($lon,$lat);
}

#
#  Transverse Mercator Projection:  scale factor
#
sub TMscale
{
   my $self = shift;
   my ($lon,$lat) = @_;

   my $k0    = $self->{K0};
   my $ee2   = $self->{ee2};
   my $lon0  = $self->{LON0};

   my $fi    = $d2r*$lat;
   my $sinfi = sin($fi);
   my $cosfi = cos($fi);
   my $tanfi = $sinfi/$cosfi;
   my $T1    = $tanfi*$tanfi;
   my $T2    = $T1*$T1;
   my $C1    = $ee2*$cosfi*$cosfi;
   my $C2    = $C1*$C1;
   my $A1    = $d2r*($lon-$lon0)*$cosfi;
   my $A2    = $A1*$A1;
   my $A4    = $A2*$A2;
   my $A6    = $A4*$A2;

   return $k0*(1+(1+$C1)*$A2/2 + (5-4*$T1+42*$C1+13*$C2-28*$ee2)*$A4/24 + (61-148*$T1+16*$T2)*$A6/720);
}

#
#  T function
#
sub T
{
   my ($sinfi,$e) = @_;
   return sqrt( (1-$sinfi)/(1+$sinfi) * ((1+$e*$sinfi)/(1-$e*$sinfi))**$e);
}

#
#  Lambert Conformal Conic Projection:  Set
#
sub LCCset
{
   my $self = shift;

   my $lat0 = $self->{LAT0};
   my $sp1  = $self->{LAT1};
   my $sp2  = $self->{LAT2};
   my $a    = $self->{a};
   my $e    = $self->{e};
   my $e2   = $self->{e2};
   (defined($sp1) && -90<$sp1 && $sp1<90) || die "Invalid standard parallel $sp1\n";
   (defined($sp2) && -90<$sp2 && $sp2<90) || die "Invalid standard parallel $sp2\n";

   my $sinfi1 = sin($d2r*$sp1);
   my $sinfi2 = sin($d2r*$sp2);
   my $m1 = cos($d2r*$sp1)/sqrt(1-$e2*$sinfi1*$sinfi1);
   my $m2 = cos($d2r*$sp2)/sqrt(1-$e2*$sinfi2*$sinfi2);
   my $t0 = T(sin($d2r*$lat0),$e);
   my $t1 = T($sinfi1,$e);
   my $t2 = T($sinfi2,$e);

   my $n  = log($m1/$m2)/log($t1/$t2);
   my $F  = $a*$m1/($n*$t1**$n);
   my $r0 = $F*$t0**$n;
   my $sign = ($n<0) ? -1 : 1;

   $self->{n}    = $n;
   $self->{F}    = $F;
   $self->{r0}   = $r0;
   $self->{sign} = $sign;
}

#
#  Lambert Conformal Conic Projection:  forward
#
sub LCCforward
{
   my $self = shift;
   my ($lon,$lat) = @_;

   my $e    = $self->{e};
   my $n    = $self->{n};
   my $F    = $self->{F};
   my $r0   = $self->{r0};
   my $lon0 = $self->{LON0};

   my $r   = $F*T(sin($d2r*$lat),$e)**$n;
   my $th  = $d2r*$n*($lon-$lon0);
   my $x   = $r*sin($th);
   my $y   = $r0 - $r*cos($th);
   return ($x,$y);
}

#
#  Lambert Conformal Conic Projection:  inverse
#
sub LCCinverse
{
   my $self = shift;
   my ($x,$y) = @_;

   my $lon0 = $self->{LON0};
   my $n    = $self->{n};
   my $e    = $self->{e};
   my $F    = $self->{F};
   my $r0   = $self->{r0};
   my $sign = $self->{sign};

   $y = $r0-$y;
   my $r = $sign*sqrt($x*$x + $y*$y);
   my $t = ($r/$F)**(1/$n);
   my $th = ($n<0) ? atan2(-$x,-$y) : atan2($x,$y);

   my $k  = 0;
   my $fi = $t;
   my $fi0 = $t-1;
   while (abs($fi-$fi0)>$tol)
   {
      ($k++ < 100) || die "LCC inverse iteration failed to converge\n";
      $fi0 = $fi;
      my $sinfi = sin($fi);
      $fi = $pi/2 - 2*atan2($t*((1-$e*$sinfi)/(1+$e*$sinfi))**($e/2),1);
   }

   my $lat = $r2d*$fi;
   my $lon = $r2d*atan2($x/$y,1)/$n + $lon0;
   return ($lon,$lat);
}

#
#  Lambert Conformal Conic Projection:  scale
#
sub LCCscale
{
   my $self = shift;
   my ($lon,$lat) = @_;

   my $e  = $self->{e};
   my $n  = $self->{n};
   my $F  = $self->{F};
   my $a  = $self->{a};
   my $e2 = $self->{e2};

   my $fi    = $d2r*$lat;
   my $sinfi = sin($fi);
   my $m     = cos($fi)/sqrt(1-$e2*$sinfi*$sinfi);
   my $r     = $F*T($sinfi,$e)**$n;
   return $r*$n/($a*$m);
}

#
#  Q function
#
sub Q
{
   my ($sinfi,$e,$e2) = @_;
   return (1-$e2)*($sinfi/(1-$e2*$sinfi*$sinfi) - 0.5/$e*log((1-$e*$sinfi)/(1+$e*$sinfi)));
}

#
#  Albers Equal-Area Conic Projection:  set
#
sub AECset
{
   my $self = shift;

   my $lat0 = $self->{LAT0};
   my $sp1  = $self->{LAT1};
   my $sp2  = $self->{LAT2};
   my $a    = $self->{a};
   my $e    = $self->{e};
   my $e2   = $self->{e2};
   (defined($sp1) && -90<$sp1 && $sp1<90) || die "Invalid standard parallel $sp1\n";
   (defined($sp2) && -90<$sp2 && $sp2<90) || die "Invalid standard parallel $sp2\n";

   my $sinfi0 = sin($d2r*$lat0);
   my $sinfi1 = sin($d2r*$sp1);
   my $sinfi2 = sin($d2r*$sp2);

   my $m1 = cos($d2r*$sp1)/sqrt(1-$e2*$sinfi1*$sinfi1);
   my $m2 = cos($d2r*$sp2)/sqrt(1-$e2*$sinfi2*$sinfi2);
   my $q0 = Q($sinfi0,$e,$e2);
   my $q1 = Q($sinfi1,$e,$e2);
   my $q2 = Q($sinfi2,$e,$e2);

   my $n = ($m1*$m1-$m2*$m2)/($q2-$q1);
   my $C = $m1*$m1+$n*$q1;
   my $r0 = $a*sqrt($C-$n*$q0)/$n;

   $self->{n}  = $n;
   $self->{C}  = $C;
   $self->{r0} = $r0;
}

#
#  Albers Equal-Area Conic Projection:  forward
#
sub AECforward
{
   my $self = shift;
   my ($lon,$lat) = @_;

   my $a    = $self->{a};
   my $n    = $self->{n};
   my $C    = $self->{C};
   my $r0   = $self->{r0};
   my $e    = $self->{e};
   my $e2   = $self->{e2};
   my $lon0 = $self->{LON0};

   my $q   = Q(sin($d2r*$lat),$e,$e2);
   my $r   = $a*sqrt($C-$n*$q)/$n;
   my $th  = $d2r*$n*($lon-$lon0);
   my $x   = $r*sin($th);
   my $y   = $r0 - $r*cos($th);
   return ($x,$y);
}

#
#  Albers Equal-Area Conic Projection:  inverse
#
sub AECinverse
{
   my $self = shift;
   my ($x,$y) = @_;

   my $a    = $self->{a};
   my $n    = $self->{n};
   my $C    = $self->{C};
   my $r0   = $self->{r0};
   my $e    = $self->{e};
   my $e2   = $self->{e2};
   my $lon0 = $self->{LON0};

   $y = $r0-$y;
   my $r = sqrt($x*$x + $y*$y);
   my $th = ($n<0) ? atan2(-$x,-$y) : atan2($x,$y);
   my $q = ($C-($r*$n/$a)*($r*$n/$a))/$n;

   my $k  = 0;
   my $fi = Math::Trig::asin(0.5*$q);
   my $fi0 = $fi-1;
   while (abs($fi-$fi0)>$tol)
   {
      ($k++ < 100) || die "AEC inverse iteration failed to converge\n";
      $fi0 = $fi;
      my $sinfi = sin($fi);
      my $es2 = 1-$e2*$sinfi*$sinfi;
      $fi += 0.5*$es2*$es2/cos($fi)*($q/(1-$e2) - $sinfi/$es2 + 0.5/$e*log((1-$e*$sinfi)/(1+$e*$sinfi)));
   }

   my $lat = $r2d*$fi;
   my $lon = $r2d*$th/$n + $lon0;
   return ($lon,$lat);
}

#
#  Albers Equal-Area Conic Projection:  scale factor
#
sub AECscale
{
   my $self = shift;
   my ($lon,$lat) = @_;

   my $n    = $self->{n};
   my $C    = $self->{C};
   my $e    = $self->{e};
   my $e2   = $self->{e2};

   my $fi    = $d2r*$lat;
   my $sinfi = sin($fi);
   my $q     = Q($sinfi,$e,$e2);
   return sqrt(($C-$n*$q)*(1-$e2*$sinfi*$sinfi))/cos($fi);
}

#
#  NAD83 shift - shared by all instances
#
my %nad;
sub NAD83set
{
   my $self = shift;
   #  nad83 file is shared by all instances
   unless ($nad{FILE})
   {
      $nad{FILE} = PM::grflib('data/nad83.dat');
      $nad{FD} = new FileHandle($nad{FILE});
      $nad{FD} || die "Cannot open $nad{FILE}\n";
      binmode($nad{FD});
      my $buf;
      (read($nad{FD},$buf,1)==1) || die "Cannot read $nad{FILE}\n";
      $nad{REV} = $buf;
      $nad{N} = GetI32($nad{FD},$nad{REV});
      my $l = 1+4+(8+6*4)*$nad{N};
      for (my $k=0;$k<$nad{N};$k++)
      {
         (read($nad{FD},$nad{NAME}[$k],8)==8) || die "Cannot read $nad{FILE}\n";;
         $nad{XMIN}[$k] = GetR32($nad{FD},$nad{REV});
         $nad{YMIN}[$k] = GetR32($nad{FD},$nad{REV});
         $nad{DX}[$k]   = GetR32($nad{FD},$nad{REV});
         $nad{DY}[$k]   = GetR32($nad{FD},$nad{REV});
         $nad{NX}[$k]   = GetI32($nad{FD},$nad{REV});
         $nad{NY}[$k]   = GetI32($nad{FD},$nad{REV});
         $nad{XMAX}[$k] = $nad{XMIN}[$k] + ($nad{NX}[$k]-1)*$nad{DX}[$k];
         $nad{YMAX}[$k] = $nad{YMIN}[$k] + ($nad{NY}[$k]-1)*$nad{DY}[$k];
         $nad{OFF}[$k] = $l;
         $l += 8*$nad{NX}[$k]*$nad{NY}[$k]; 
      }
   }
}
sub NAD83shift
{
   my $self = shift;
   my ($lon,$lat) = @_;
   my $k;
   for (my $l=0;!$k && $l<$nad{N};$l++)
   {
      ($nad{XMIN}[$l]<=$lon && $lon<=$nad{XMAX}[$l] && $nad{YMIN}[$l]<=$lat && $lat<=$nad{YMAX}[$l]) && ($k = $l);
   }
   defined($k) || die "NAD83 (lon,lat) out of range: ($lon,$lat)\n";

   my $i = int(($lon-$nad{XMIN}[$k])/$nad{DX}[$k]);
   my $j = int(($lat-$nad{YMIN}[$k])/$nad{DY}[$k]);
   ($i==$nad{NX}[$k]) && ($i--);
   ($j==$nad{NY}[$k]) && ($j--);
   my $x0 = $i*$nad{DX}[$k]+$nad{XMIN}[$k];
   my $y0 = $j*$nad{DY}[$k]+$nad{YMIN}[$k];

   seek($nad{FD},8*($i+$j*$nad{NX}[$k])+$nad{OFF}[$k],0) || die "Cannot seek $nad{FILE}\n";
   my $x00 = GetR32($nad{FD},$nad{REV});
   my $y00 = GetR32($nad{FD},$nad{REV});
   my $x10 = GetR32($nad{FD},$nad{REV});
   my $y10 = GetR32($nad{FD},$nad{REV});

   seek($nad{FD},8*($i+($j+1)*$nad{NX}[$k])+$nad{OFF}[$k],0) || die "Cannot seek $nad{FILE}\n";
   my $x01 = GetR32($nad{FD},$nad{REV});
   my $y01 = GetR32($nad{FD},$nad{REV});
   my $x11 = GetR32($nad{FD},$nad{REV});
   my $y11 = GetR32($nad{FD},$nad{REV});

   my $fx = ($lon-$x0)/$nad{DX}[$k];
   my $fy = ($lat-$y0)/$nad{DY}[$k];
   my $xs = (1-$fx)*(1-$fy)*$x00 + $fx*(1-$fy)*$x10 + (1-$fx)*$fy*$x01 + $fx*$fy*$x11;
   my $ys = (1-$fx)*(1-$fy)*$y00 + $fx*(1-$fy)*$y10 + (1-$fx)*$fy*$y01 + $fx*$fy*$y11;
   return ($xs,$ys);
}
sub NAD83forward
{
   my $self = shift;
   my ($lon,$lat) = @_;
   my ($xs,$ys) = $self->NAD83shift($lon,$lat);
   return ($lon+$xs/3600,$lat+$ys/3600);
}
sub NAD83inverse
{
   my $self = shift;
   my ($lon,$lat) = @_;
   my ($xs,$ys) = $self->NAD83shift($lon,$lat);
   return ($lon-$xs/3600,$lat-$ys/3600);
}

#
#  Set up anisotropic transformation
#
sub xfm3
{
   my ($X,$Y,$x,$y) = @_;
   my @xfm;
   foreach my $k (0,1)
   {
      my $A = [[$X->[0],$Y->[0],1] , [$X->[1],$Y->[1],1] , [$X->[2],$Y->[2],1]];
      my $b = $k ? [$y->[0],$y->[1],$y->[2]] : [$x->[0],$x->[1],$x->[2]];
      lusolve($A,$b) || die "Cannot invert xfm3 matrix\n";
      $xfm[$k] = $b;
   }
   return \@xfm;
}

#
#  Apply anisotropic transformation
#
sub xfm3ap
{
   my ($xfm,$X,$Y) = @_;
   my $x = $xfm->[0][0]*$X + $xfm->[0][1]*$Y + $xfm->[0][2];
   my $y = $xfm->[1][0]*$X + $xfm->[1][1]*$Y + $xfm->[1][2];
   return ($x,$y);
}

#
#  Solve linear system in place
#  Note that this destroys both A and x
#
sub lusolve
{
   my ($A,$x) = @_;
   my @w;
   my $n = @$A;

   #  Decompose
   for (my $i=0;$i<$n;$i++)
   {
      my $Amax = 0;
      for (my $j=0;$j<$n;$j++)
      {
         ($Amax < abs($A->[$i][$j])) && ($Amax = abs($A->[$i][$j]));
      }
      $Amax || die "Amax==0\n";
      $w[$i] = 1/$Amax;
   }
   for (my $j=0;$j<$n;$j++)
   {
      for (my $i=0;$i<$j;$i++)
      {
         my $sum = $A->[$i][$j];
         for (my $k=0;$k<$i;$k++)
         {
            $sum -= $A->[$i][$k] * $A->[$k][$j];
         }
         $A->[$i][$j] = $sum;
      }
      my $I = $j;
      my $D = 0;
      for (my $i=$j;$i<$n;$i++)
      {
         my $sum = $A->[$i][$j];
         for (my $k=0;$k<$j;$k++)
         {
            $sum -= $A->[$i][$k] * $A->[$k][$j];
         }
         $A->[$i][$j] = $sum;
         if ($w[$i]*abs($sum) > $D)
         {
            $I  = $i;
            $D = $w[$i]*abs($sum);
         }
      }
      if ($j!=$I)
      {
         for (my $k=0;$k<$n;$k++)
         {
            my $tmp      = $A->[$I][$k];
            $A->[$I][$k] = $A->[$j][$k];
            $A->[$j][$k] = $tmp;
         }
         my $tmp  = $x->[$I];
         $x->[$I] = $x->[$j];
         $x->[$j] = $tmp;
         $w[$I] = $w[$j];
      }
      $A->[$j][$j] || die "A(j,j)==0\n";
      my $tmp = 1/$A->[$j][$j];
      for (my $i=$j+1;$i<$n;$i++)
      {
         $A->[$i][$j] *= $tmp;
      }
   }

   #  Backsolve
   for (my $i=0;$i<$n;$i++)
   {
      for (my $j=0;$j<$i;$j++)
      {
         $x->[$i] -= $A->[$i][$j] * $x->[$j];
      }
   }
   for (my $i=$n-1;$i>=0;$i--)
   {
      for (my $j=$i+1;$j<$n;$j++)
      {
         $x->[$i] -= $A->[$i][$j] * $x->[$j];
      }
      $x->[$i] /= $A->[$i][$i];
   }
   return 1;
}

#
#  Interpolate to Quarters
#
sub FindQuarter
{
   my ($fd,$KC,$cnr,@IC) = @_;
   #  Read Quarters
   my %xy;
   foreach my $Q ('A'..'D')
   {
      $xy{X}{$Q} = GetR32($fd,'I');
      $xy{Y}{$Q} = GetR32($fd,'I');
   }
   close($fd);
   #  Interpolate to quarters
   my $Qf = 0.5;
   my $Qg = 0.5;
   my $Qq = 0.5;
   for (my $k=$KC;$k<@IC;$k++)
   {
      $Qq /= 2;
      $Qf += ($IC[$k]==1 || $IC[$k]==4) ? +$Qq : -$Qq;
      $Qg += ($IC[$k]==1 || $IC[$k]==2) ? +$Qq : -$Qq;
   }
   #  Adjust to corners
   ($cnr =~ /[AD]/) && ($Qf += $Qq);
   ($cnr =~ /[BC]/) && ($Qf -= $Qq);
   ($cnr =~ /[AB]/) && ($Qg += $Qq);
   ($cnr =~ /[CD]/) && ($Qg -= $Qq);
   #  Interpolate
   my $Pf = 1-$Qf;
   my $Pg = 1-$Qg;
   my $X = $Qg*$Qf*$xy{X}{A} + $Qg*$Pf*$xy{X}{B} + $Pg*$Pf*$xy{X}{C} + $Pg*$Qf*$xy{X}{D};
   my $Y = $Qg*$Qf*$xy{Y}{A} + $Qg*$Pf*$xy{Y}{B} + $Pg*$Pf*$xy{Y}{C} + $Pg*$Qf*$xy{Y}{D};
   return ($X,$Y);
}

#
#  Find section
#
sub FindSection
{
   my ($PM,$KTRS,$cnr,@IC) = @_;
   my $S = $KTRS%64;
   $KTRS = int($KTRS/64);
   my $file = PM::grflib("data/$PM.gcdb");
   my $fd;
   open($fd , "<$file") || die "Cannot open file $file\n";
   binmode($fd);
   my $n = GetI32($fd,'I');
   my $LD = 0;
   my ($LD0,$ID);
   for (my $k=0;$k<$n;$k++)
   {
      $LD0 = $LD;
      $ID = GetI32($fd,'I');
      $LD = GetI32($fd,'I');
      ($ID == $KTRS) && last;
   }
   ($ID != $KTRS) && return;

   #  Find sections
   $n = ($LD-$LD0)/34;
   while (@IC>2) {pop @IC;}
   for (my $KC=@IC;$KC>=0;$KC--)
   {
      seek($fd,$LD0,SEEK_SET) || die "Cannot seek PM file\n";
      my $KD = $S;
      ($KC>=1) && ($KD +=   256*$IC[0]);
      ($KC>=2) && ($KD += 8*256*$IC[1]);
      for (my $l=0;$l<$n;$l++)
      {
         $ID = GetI16($fd,'I');
         if ($ID==$KD)
         {
            return FindQuarter($fd,$KC,$cnr,@IC);
         }
         else
         {
            seek($fd,32,SEEK_CUR) || die "Cannot skip PM file\n";
         }
      }
   }
   close($fd);
   return;
}

#
#  Find adhoc
#
sub FindAdhoc
{
   my ($PM,$KTRS,$cnr,@IC) = @_;
   #  Encode point
   my $L=0;
   for (my $i=@IC-1;$i>=0;$i--)
   {
      $L = 4*$L + $IC[$i]-1;
   }
   $L = 16*$L+2*@IC;

   #  Read database
   my $file = PM::grflib('data/adhoc.gcdb');
   my ($fd,$buf);
   open($fd , "<$file") || die "Cannot open file $file\n";
   binmode($fd);
   while (read($fd,$buf,8)==8)
   {
      my ($m,$k,$l) = unpack('CVV',"$buf\0");
      #  Point data
      unless ($l&1)
      {
         read($fd,$buf,8);
         ($m == $PM && $k==$KTRS && $l==$L) || next;
         close($fd);
         return unpack('f*',$buf);
      }
      #  Corner data
      {
         read($fd,$buf,24);
         ($m == $PM && $k==$KTRS) || next;
         my $KC = int($l/2)&7;
         $l = int($l/16);
         for ($k=0;$k<$KC;$k++)
         {
            (($l&3)==$IC[$k]-1) || last;
            $l = int($l/4);
         }
         ($k==$KC) || next;
         return FindQuarter($fd,$KC,$cnr,@IC);
      }
   }
   close($fd);
   return;
}

#
#  Get Lon/Lat from Township-Range-Section
#
my %IM;
sub EncodeTRS
{
   my ($trs) = @_;
   #  Read PM first time
   unless (%IM)
   {
      my $file = PM::grflib('data/pm.dat');
      my $fd;
      open($fd , "<$file") || die "Cannot open file $file\n";
      <$fd>;
      while (my $line = <$fd>)
      {
         chomp $line;
         ($line =~ /^\s*END/i) && last;
         my ($im,$pm) = split(' ',$line);
         my @pm = split(':',$pm);
         foreach $pm (@pm)
         {
            $IM{$pm} = $im;
         }
      }
      close($fd);
   }
   #  Decode TRS
   my $k = (substr($trs,2,1) =~ /[0-9]/) ? 1 : 2;
   my $PM = substr($trs,0,$k);
   exists($IM{$PM}) || die "Unknown PM $PM\n";
   $PM = $IM{$PM};
   my $CQ = substr($trs,$k++,1);
   my $T  = substr($trs,$k,3); $k += 3;
   my $Tq = (substr($trs,$k,1) =~ /[QHT]/i) ? uc substr($trs,$k++,1) : '';
   my $R  = substr($trs,$k,3); $k += 3;
   my $Rq = (substr($trs,$k,1) =~ /[QHT]/i) ? uc substr($trs,$k++,1) : '';
   my $S  = substr($trs,$k,2); $k += 2;
   my $Sq = (substr($trs,$k,1) =~ /\*/) ? substr($trs,$k++,1) : '';
   #  Only regular A-D for now
   my @IC;
   for (;$k<length($trs);$k++)
   {
      my $Q = substr($trs,$k,1);
      ($Q =~ /[A-D]/) || last;
      push @IC , ord($Q)-64;
   }
   my $cnr = '';
   if ($k<length($trs) && substr($trs,$k,1)eq'>')
   {
      $k++;
      (@IC>0) || die "Invalid corner\n";
      $cnr = chr(64+pop @IC);
   }
   #  Encode TRS
   my %off = (''=>0,Q=>1,H=>2,T=>3);
   $T = ($CQ=~/[AB]/?1024:0) + 4*$T + $off{$Tq};
   $R = ($CQ=~/[AD]/?1024:0) + 4*$R + $off{$Rq};
   my $KTRS = 262144*$T+128*$R+64*($Sq?1:0)+$S;
   return ($PM,$KTRS,$cnr,@IC);
}

sub TRS2XY
{
   my ($trs) = @_;
   my ($PM,$KTRS,$cnr,@IC) = EncodeTRS($trs);
   my ($x,$y) = FindSection($PM,$KTRS,$cnr,@IC);
   defined($x) || (($x,$y) = FindAdhoc($PM,$KTRS,$cnr,@IC));
   return ($x,$y);
}

1;
